package com.beechwood.certificates;

public class ASN1Value {
  private int asn1Class;
  private boolean constructed = false;
  private int tag;
  private int length;
  protected byte[] data;
  private int totalBytes;

  public ASN1Value(byte[] array, int startPosition)  {
    byte b = array[startPosition];
    asn1Class = b & ASN1Constants.MASK_CLASS;
    if ((b & ASN1Constants.BIT_6) != 0)
      constructed = true;
    int tagLength = 1;
    tag = array[startPosition] & ASN1Constants.BITS_5_1;
    if (tag == ASN1Constants.BITS_5_1)
      tagLength = decodeHighTag(array, startPosition + 1);
    int p = startPosition + tagLength;
    b = array[p];
    if ((b & ASN1Constants.BIT_8) == 0) {
      length = ((int)b) & ASN1Constants.BITS_7_1;
    }
    else {
      length = 0;
      for (int j = 0; j < (((int)b) & ASN1Constants.BITS_7_1); ++j) {
        ++p;
        length = (length << 8) | ((array[p] & ASN1Constants.BITS_8_1));
      }
    }
    if (length == 0) {
      length = -(array.length - p - 1);
    }
    ++p;
    totalBytes = p - startPosition + Math.abs(length);
    data = new byte[Math.abs(length)];
    System.arraycopy(array,p,data,0,data.length);
  }

  private int decodeHighTag(byte[] array, int startPosition) {
    int tag = 0;
    int i = startPosition;
    boolean h = true;
    while (true) {
      byte b = array[i++];
      tag = (tag << 7) + (((int)b) & ASN1Constants.BITS_7_1);
      if  ((b & ASN1Constants.BIT_8) == 0) {
        break;
      }
    }
    return i - startPosition;
  }

  public boolean hasIndefiniteLength() {
    return (length < 0);
  }

  public boolean isConstructed() {
    return constructed;
  }

  public byte[] data() {
    return data;
  }

  public String dataAsHexString() {
    StringBuffer sb = new StringBuffer();
    for (int i = 1; i < data.length; ++i) {
      int j = data[i] & ASN1Constants.BITS_8_1;
      if (j < 16)
        sb.append("0");
      sb.append(Integer.toHexString(j) + ":");
    }
    sb.setLength(sb.length() - 1);
    return sb.toString();
  }

  public int totalBytes() {
    return totalBytes;
  }

  public int getASN1Class() {
    return asn1Class;
  }

  public int tag() {
    return tag;
  }

  public int getInteger() throws ASN1ValueException {
    if (tag != ASN1Constants.ID_INTEGER)
      throw new ASN1ValueException("getInteger - not INTEGER");
    int val = 0;
    for (int i = 0; i < data.length; ++i)
      val = (val << 8) + (data[i] & ASN1Constants.BITS_8_1);
    return val;
  }

  public int length() {
    return length;
  }

  public String getOID() throws ASN1ValueException {
    if (tag != ASN1Constants.ID_OBJECT_IDENTIFIER)
      throw new ASN1ValueException("not OID");
    StringBuffer sb = new StringBuffer();
    int b = data[0] & ASN1Constants.BITS_8_1;
    sb.append(Integer.toString(b / 40));
    sb.append(".");
    sb.append(Integer.toString(b % 40));
    int len = 0;
    for (int i = 1; i < data.length; ++i) {
      b = data[i] & ASN1Constants.BITS_8_1;
      if ((b & ASN1Constants.BIT_8) == ASN1Constants.BIT_8) {
        len = (len << 7) + (b & ASN1Constants.BITS_7_1);
      }
      else {
        len = (len << 7) + (b & ASN1Constants.BITS_8_1);
        sb.append(".");
        sb.append(Integer.toString(len));
        len = 0;
      }
    }
    return sb.toString();
  }

  public String getOIDDescription() throws X509CertificateException {
    String oid = null;
    try {
      oid = getOID();
    }
    catch (ASN1ValueException e) {
      throw new X509CertificateException(e.getMessage());
    }
    if (ASN1Constants.oidMap.containsKey(oid))
      return (String)(ASN1Constants.oidMap.get(oid));
    else
      return("UNKNOWN");
  }

  public String getString() {
    return new String(data);
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("ASN1Value [class = ");
    sb.append(Integer.toString(asn1Class));
    sb.append(", constructed = ");
    sb.append(new Boolean(constructed).toString());
    sb.append(", tag = ");
    sb.append(Integer.toString(tag));
    sb.append(", length = ");
    sb.append(Integer.toString(length));
    sb.append(", totalBytes = ");
    sb.append(Integer.toString(totalBytes));
    if (data.length > 0) {
      sb.append(", first ");
      int dbl = (data.length < 10) ? data.length : 10;
      sb.append(Integer.toString(dbl));
      sb.append(" data bytes: ");
      for (int i = 0; i < Math.min(data.length,10); ++i) {
        int j = data[i] & ASN1Constants.BITS_8_1;
        if (j < 16)
          sb.append("0");
        sb.append(Integer.toHexString(j) + ":");
      }
      sb.append("]");
    }
    return new String(sb);
  }
}
